# ============================================================================
# =======                    TEH Doorbell Software                     =======
# =======                 Version 2.6.1  |  06/05/2009                 =======
# =======            By: Nick Viera |  http://www.tehhouse.us/         =======
# =======               Copyright (c) 2009 Nick Viera                  =======
# ============================================================================
#
#   * This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   * This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   * You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ============================================================================
#
#  Timer class code based on that posted by James Stroud, 17 Dec 2005.
#  See http://code.activestate.com/recipes/464959/ for more information.
#
# ============================================================================


# ===== Import modules needed for using extended features of Python ==========
import serial
import string
import pyaudio
import wave
import sys
import time
import datetime
import threading
import os
import Image
import smtplib
from   Tkinter  import *

# ===== Global variables for serial port =====================================
ser         = serial.Serial()
state       = False
state_DSR   = ""
state_CTS   = ""
state_RI    = ""
state_CD    = ""
state_RTS   = ""

# ===== Global variables for general app functions ===========================
app             = ""
log             = ""
status          = "Uninitialized"
status_led_hits = 0
play_sound      = 1
play_beep       = 1
send_emails     = 0

# ===== Global variables for the Tkinter GUI layout ==========================
row_checkbox    = 1
row_soundfile   = 2
row_stat_lines  = row_soundfile + 2
row_buttons     = row_stat_lines + 7
row_log         = row_buttons + 2

# ===== Load settings from file settings.py ==================================
execfile('settings.py')

class ResettableTimer(threading.Thread):
  """
  The ResettableTimer class is a timer whose counting loop can be reset
  arbitrarily. Its duration is configurable. Commands can be specified
  for both expiration and update. Its update resolution can also be
  specified. Resettable timer keeps counting until the "run" method
  is explicitly killed with the "kill" method.
  """
  def __init__(self, maxtime, expire, inc=None, update=None):
    """
    @param maxtime: time in seconds before expiration after resetting
                    in seconds
    @param expire: function called when timer expires
    @param inc: amount by which timer increments before
                updating in seconds, default is maxtime/2
    @param update: function called when timer updates
    """
    self.maxtime = maxtime
    self.expire = expire
    if inc:
        self.inc = inc
    else:
        self.inc = maxtime/2
    if update:
        self.update = update
    else:
        self.update = lambda c : None
    self.counter = 0
    self.active = True
    self.stop = False
    threading.Thread.__init__(self)
    self.setDaemon(True)
  def set_counter(self, t): self.counter = t
  def deactivate(self): self.active = False
  def kill(self): self.stop = True
  def reset(self):
    """
    Fully rewinds the timer and makes the timer active, such that
    the expire and update commands will be called when appropriate.
    """
    self.counter = 0
    self.active = True

  def run(self):
    """
    Run the timer loop.
    """
    while True:
      self.counter = 0
      while self.counter < self.maxtime:
        self.counter += self.inc
        time.sleep(self.inc)
        if self.stop:
          return
        if self.active:
          self.update(self.counter)
      if self.active:
        self.expire()
        self.active = True

# ===== Timer Events =========================================================
# Events which occur each time the timer expires. This is used to poll the
# doorbell button to see if it has been pressed since the last check.
def timer_events():
    global ser
    global state
    global status
    global state_DSR
    global state_CTS
    global state_RI
    global state_CD
    global play_sound
    global play_beep
    global sound_cmd

    if (ser.getCTS()) == True: state_CTS = "True"
    else:                      state_CTS = "False"

    if (ser.getRI()) == True:  state_RI = "True"
    else:                      state_RI = "False"

    if (ser.getCD()) == True:  state_CD = "True"
    else:                      state_CD = "False"

    if ((state == True) | (ser.getDSR() == True)):
        write_log("Door Bell Is Ringing!")  # Print to the log
        app.status_canvas.itemconfig (app.status_led, fill="yellow")
        state_DSR = "True"
        status = "Ringing!"

        # Beep in the terminal if enabled by the checkbox
        if play_beep == 1: print "\x07"

        led_off()                           # Turn off LED

        # Send an E-mail if enabled by the checkbox
        if send_emails == 1: send_email()

        # Play the sound file if enabled by the checkbox
        if play_sound == 1: os.system(sound_cmd)

        # Beep in the terminal if enabled by the checkbox
        if play_beep == 1: print "\x07"

        state = False
        led_on()                            # Turn on LED
    else:
        state_DSR = "False"
        status = "Polling"
    app.update_vars()

def led_on():
    global state_RTS
    ser.setRTS(1)    # Turn on LED
    state_RTS = "True"
    app.update_vars()

def led_off():
    global state_RTS
    global ser
    openport = ser.isOpen()
    if openport == True:
        ser.setRTS(0)    # Turn off LED
        state_RTS = "False"
        app.update_vars()

class rtTester(Frame):
    global state
    global state_DSR
    def __init__(self, master=None):
        # Start Timer !
        self.timer = ResettableTimer(0.1, self.command, inc=0.05, update=self.update)
        Frame.__init__(self, master)    # New top-level frame w/ no master
        top = self.winfo_toplevel()
        top.rowconfigure(0, weight=1)
        top.columnconfigure(0, weight=1)
        self.grid(sticky=N+W, ipadx=10, ipady=10) # Make the frame visible on grid 
        self.make_vars()
        self.make_widgets()
        self.timer.start()
        self.timer.deactivate()

    def destroy(self, *args, **kwargs):
        self.timer.kill()
        Frame.destroy(self, *args, **kwargs)

    def command(self, e=None):
        self.timer.reset()
        timer_events()

    def update(self, c):
        global status_led_hits
        if status_led_hits <= 3:
                self.status_canvas.itemconfig (self.status_led, fill="gray")
        elif 3 < status_led_hits < 6:
                self.status_canvas.itemconfig (self.status_led, fill="green")
        else:   status_led_hits=0
        status_led_hits = (status_led_hits + 1)

    def rewind(self, e=None):
        global state
        self.timer.reset()
        state = True

    def deactivate(self, e=None):
        global status
        status = "Stopped"
        write_log("Polling Stopped.")
        self.timer.deactivate()
        led_off()
        self.status_canvas.itemconfig (self.status_led, fill="red")
        self.ringbut.grid_remove()
        self.ringbut = Button(self, text='Ring', state=DISABLED, command=self.ring)
        self.ringbut.grid(sticky=W+E, column=3, row=row_buttons, padx=2)
        self.update_vars()
        serial_close()

    def reactivate(self, e=None):
        serial_open()
        self.timer.reset()
        led_on()
        write_log("Polling Started")
        self.ringbut.grid_remove()
        self.ringbut = Button(self, text='Ring', state=NORMAL, command=self.ring)
        self.ringbut.grid(sticky=W+E, column=3, row=row_buttons, padx=2)
        self.update_vars()

    def make_vars(self):
        self.gstate_DSR = StringVar()
        self.gstate_CTS = StringVar()
        self.gstate_RI  = StringVar()
        self.gstate_CD  = StringVar()
        self.gstatus    = StringVar()
        self.gstate_RTS = StringVar()
        self.glog       = StringVar()

    def ring(self):
        global state
        state = True

    def reset(self):
        self.deactivate()
        clear_log()

    def update_vars(self):
        global state_DSR
        global state_CTS
        global state_RI
        global state_CD
        global status
        global state_RTS
        global log

        self.gstate_DSR.set(state_DSR)
        if state_DSR == "True": self.DSR_canvas.itemconfig (self.DSR_led, fill="green")
        else:                   self.DSR_canvas.itemconfig (self.DSR_led, fill="red")

        self.gstate_CTS.set(state_CTS)
        if state_CTS == "True": self.CTS_canvas.itemconfig (self.CTS_led, fill="green")
        else:                   self.CTS_canvas.itemconfig (self.CTS_led, fill="red")

        self.gstate_RI.set(state_RI)
        if state_RI == "True":  self.RI_canvas.itemconfig (self.RI_led, fill="green")
        else:                   self.RI_canvas.itemconfig (self.RI_led, fill="red")

        self.gstate_CD.set(state_CD)
        if state_CD == "True":  self.CD_canvas.itemconfig (self.CD_led, fill="green")
        else:                   self.CD_canvas.itemconfig (self.CD_led, fill="red")

        self.gstatus.set(status)
        self.gstate_RTS.set(state_RTS)
        if state_RTS == "True": self.RTS_canvas.itemconfig (self.RTS_led, fill="green")
        else:                   self.RTS_canvas.itemconfig (self.RTS_led, fill="red")

        self.glog.set(log)

    def toggle_sound_file(self):
        global play_sound
        if play_sound == 1: play_sound=0
        else:               play_sound=1

    def toggle_beep(self):
        global play_beep
        if play_beep == 1: play_beep=0
        else:              play_beep=1

    def toggle_email(self):
        global send_emails
        if send_emails == 1: send_emails=0
        else:                send_emails=1

    def make_widgets(self):
        horizline_leng  = 400

        global row_buttons
        global row_checkbox
        global row_soundfile
        global row_stat_lines
        global row_log


        #== Horizontal Line =========================================
        self.line_canvas =    Canvas (self, height=2)
        self.line_line   =    self.line_canvas.create_line(0,1,horizline_leng,1, width=2,fill="#777777")
        self.line_canvas.grid (column=0, row=(row_checkbox-1), columnspan=7, sticky=N+E+W+S, pady=4)

        self.checkbox_sound =       Checkbutton (self, height=1, text="Play Sound File", command=self.toggle_sound_file)
        self.checkbox_sound.select()
        self.checkbox_sound.grid    (column=0, columnspan=2, row=row_checkbox, pady=4, sticky=N+W)

        self.checkbox_beep =       Checkbutton (self, height=1, text="System Bell", command=self.toggle_beep)
        self.checkbox_beep.select()
        self.checkbox_beep.grid    (column=2, columnspan=1, row=row_checkbox, pady=4, sticky=N+W)

        self.checkbox_email =       Checkbutton (self, height=1, text="Send E-mail", command=self.toggle_email)
        self.checkbox_email.deselect()
        self.checkbox_email.grid    (column=3, columnspan=2, row=row_checkbox, pady=4, sticky=N+W)

        self.soundfile_label =      Label(self,text="Sound (WAV) File:  ", height=1)
        self.soundfile_label.grid   (row=row_soundfile, column=0, columnspan=2)
        self.soundfile_input =      Text (self,height=1,width=25)
        self.soundfile_input.grid   (row=row_soundfile, column=2, columnspan=3, sticky=E+W)

        #== Horizontal Line =========================================
        self.line_canvas =    Canvas (self, height=2)
        self.line_line   =    self.line_canvas.create_line(0,1,horizline_leng,1, width=2,fill="#777777")
        self.line_canvas.grid (column=0, row=(row_soundfile+1), columnspan=7, sticky=N+E+W+S, pady=4)

        #== Status Line Widgets =====================================
        self.status_canvas =    Canvas (self, width=12, height=12)
        self.status_led_bg =    self.status_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.status_led =       self.status_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="red")
        self.status_canvas.grid (column=0, row=row_stat_lines, padx=5)

        self.status_label =     Label(self, text='Status:', fg='black')
        self.status_label.grid  (sticky=W, column=1, row=row_stat_lines)
        self.status_value =     Label(self, text='?', fg='blue', textvariable=self.gstatus)
        self.status_value.grid  (sticky=W, column=2, row=row_stat_lines, columnspan=2)

        #== Logo =================================================
        self.logo_file =             PhotoImage(file='tehlogo.gif')
        self.logo_canvas =      Canvas(self, width=115, height=125)
        self.logoimg =          self.logo_canvas.create_image(5,1,image=self.logo_file,anchor=NW)
        self.logotext =         self.logo_canvas.create_text(1,90,anchor=NW,justify=CENTER,text="Nick Viera\nVer. 2.6 - 06/2009", font=("arial","10","bold"))
        self.logo_canvas.grid   (row=row_stat_lines, column=3, columnspan=4, rowspan=6, sticky=E)

        #== DSR Line Widgets =====================================
        self.DSR_canvas =       Canvas (self, width=12, height=12)
        self.DSR_led_bg =       self.DSR_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.DSR_led =          self.DSR_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="green")
        self.DSR_canvas.grid    (column=0, row=(row_stat_lines+1))

        self.text_tDSR =        Label(self, text='DSR line:', fg='black')
        self.text_tDSR.grid     (sticky=W, column=1, row=(row_stat_lines+1))
        self.text_DSR =         Label(self, text='?', fg='blue', textvariable=self.gstate_DSR)
        self.text_DSR.grid      (sticky=W, column=2, row=(row_stat_lines+1), columnspan=2)

        #== RTS Line Widgets =====================================
        self.RTS_canvas =       Canvas (self, width=12, height=12)
        self.RTS_led_bg =       self.RTS_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.RTS_led =          self.RTS_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="green")
        self.RTS_canvas.grid    (column=0, row=(row_stat_lines+2))

        self.text_tRTS = Label(self, text='RTS line:', fg='black')
        self.text_tRTS.grid(sticky=W, column=1, row=(row_stat_lines+2))
        self.text_RTS = Label(self, text='?', fg='blue', textvariable=self.gstate_RTS)
        self.text_RTS.grid(sticky=W, column=2, row=(row_stat_lines+2), columnspan=2)

        #== CTS Line Widgets =====================================
        self.CTS_canvas =       Canvas (self, width=12, height=12)
        self.CTS_led_bg =       self.CTS_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.CTS_led =          self.CTS_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="green")
        self.CTS_canvas.grid    (column=0, row=(row_stat_lines+3))

        self.text_tCTS = Label(self, text='CTS line:', fg='black')
        self.text_tCTS.grid(sticky=W, column=1, row=(row_stat_lines+3))
        self.text_CTS = Label(self, text='?', fg='blue', textvariable=self.gstate_CTS)
        self.text_CTS.grid(sticky=W, column=2, row=(row_stat_lines+3), columnspan=2)

        #== RI Line Widgets =====================================
        self.RI_canvas =       Canvas (self, width=12, height=12)
        self.RI_led_bg =       self.RI_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.RI_led =          self.RI_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="green")
        self.RI_canvas.grid    (column=0, row=(row_stat_lines+4))

        self.text_tRI = Label(self, text='RI line:', fg='black')
        self.text_tRI.grid(sticky=W, column=1, row=(row_stat_lines+4))
        self.text_RI = Label(self, text='?', fg='blue', textvariable=self.gstate_RI)
        self.text_RI.grid(sticky=W, column=2, row=(row_stat_lines+4), columnspan=2)

        #== CD Line Widgets =====================================
        self.CD_canvas =       Canvas (self, width=12, height=12)
        self.CD_led_bg =       self.CD_canvas.create_oval(1,1,12,12,outline="#555555",fill="#777777")
        self.CD_led =          self.CD_canvas.create_oval(3,3,10,10,outline="#aaaaaa",fill="green")
        self.CD_canvas.grid    (column=0, row=(row_stat_lines+5))

        self.text_tCD = Label(self, text='CD line:', fg='black')
        self.text_tCD.grid(sticky=W, column=1, row=(row_stat_lines+5))
        self.text_CD = Label(self, text='?', fg='blue', textvariable=self.gstate_CD)
        self.text_CD.grid(sticky=W, column=2, row=(row_stat_lines+5), columnspan=2)

        #== Horizontal Line =========================================
        self.line_canvas =    Canvas (self, height=2)
        self.line_line   =    self.line_canvas.create_line(0,1,horizline_leng,1, width=2,fill="#777777")
        self.line_canvas.grid (column=0, row=(row_stat_lines+6), columnspan=7, sticky=N+E+W+S, pady=4)

        self.rwbut = Button(self, text='Reset', command=self.reset)
        self.rwbut.grid(sticky=W+E, column=4, columnspan=3, row=row_buttons, padx=2)
        self.ringbut = Button(self, text='Ring', state=DISABLED, command=self.ring)
        self.ringbut.grid(sticky=W+E, column=3, row=row_buttons, padx=2)
        self.deactbut = Button(self, text='Stop Polling', command=self.deactivate)
        self.deactbut.grid(sticky=W+E, column=2, row=row_buttons, padx=2)
        self.reactbut = Button(self, text='Start Polling', command=self.reactivate, state=NORMAL)
        self.reactbut.grid(sticky=W+E, column=0, row=row_buttons, columnspan=2, padx=2)

        #== Horizontal Line =========================================
        self.line_canvas =    Canvas (self, height=2)
        self.line_line   =    self.line_canvas.create_line(0,1,horizline_leng,1, width=2,fill="#777777")
        self.line_canvas.grid (column=0, row=(row_buttons+1), columnspan=7, sticky=N+E+W+S, pady=4)

        self.text_tlog = Label(self, text='Event Log:', fg='black')
        self.text_tlog.grid(sticky=W, column=0, columnspan=2, row=row_log)
        self.text_log = Text(self, fg='black', bg='white', border=2, relief="ridge", height=5, width=39, wrap="none")
        self.text_log.grid(sticky=W+E+N+S, column=0, row=(row_log+1), columnspan=6)

        self.scrollY = Scrollbar ( self, orient=VERTICAL, command=self.text_log.yview )
        self.scrollY.grid ( row=(row_log+1), column=6, sticky=N+S )
        self.scrollX = Scrollbar ( self, orient=HORIZONTAL, command=self.text_log.xview )
        self.scrollX.grid ( row=(row_log+2), column=0, sticky=E+W, columnspan=6 )

        self.text_log["xscrollcommand"]  =  self.scrollX.set
        self.text_log["yscrollcommand"]  =  self.scrollY.set

def send_email():
    global smtp_server
    global smtp_auth
    global smtp_user
    global smtp_pass
    global email_to
    global email_from
    global email_body
    global email_subject
    global email_msg
    global smtp_port

    email_msg = ("From: %s\r\nTo: %s\r\nSubject: %s\r\n\r\n%s" % (email_from, email_from, email_subject, email_body))

    # Connect to the SMTP server at address and port given
    session = smtplib.SMTP(smtp_server, smtp_port)

    # If authentication is required, login with a username and password
    if smtp_auth == 1: session.login(smtp_user, smtp_pass)

    smtpresult = session.sendmail(email_from, email_to, email_msg)

    # smtpresult will be non-empty if errors occurred during sending
    if smtpresult:
        write_log("Notification E-mail sending FAILED.")
        errstr = ""
        for recip in smtpresult.keys():
            errstr = """Could not delivery mail to: %s
            Server said: %s
            %s
            %s""" % (recip, smtpresult[recip][0], smtpresult[recip][1], errstr)
        raise smtplib.SMTPException, errstr

    else:
        write_log("Notification E-mail sent successfully.")

    session.quit()  # Close connection to the SMTP server.

# ===== Serial Port initialization ===========================================
# Configure the serial object in preparation for opening the serial port
def serial_init():
        global ser
        global serial_portname
        global serial_baud
        global serial_bits
        global serial_stop

        ser.port=serial_portname
        ser.baudrate=serial_baud
        ser.bytesize=serial_bits
        ser.stopbits=serial_stop
        ser.parity=serial.PARITY_NONE
        ser.timeout=5

# ===== Open Serial Port =====================================================
# Actually open the serial port. Report success or failure of the operation
def serial_open():
        global text_status
        global ser
        openport = ser.isOpen()
        if openport == True: 
            write_log("Serial port is already open!")
        else:
            ser.open()
            write_log("Serial port " + ser.portstr + " opened.")
            ser.setDTR(1)    # Setup for button
            led_on()    # Turn on LED

def clear_log():
    app.text_log.delete(1.0, END)

def write_log(texty):
    global log

    datenow = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime())
    log = datenow + " = " + texty + "\n"
    app.text_log.insert(CURRENT, log)

# ===== Close Serial Port ====================================================
# Actually close the serial port. Report success or failure of the operation
def serial_close():
        global text_status
        global ser
        openport = ser.isOpen()
        if openport == False:
            write_log("Serial port is already closed!")
        else:
            app.update_vars()
            led_off()    # Turn off LED
            ser.close()
            write_log("Serial port " + ser.portstr + " closed.")

# ===== Top level function ===================================================
# Intitialize a new GUI object using Tkinter and create a new master frame
# this is where the entire program starts and becomes visible to the user
def test():
    global app
    app = rtTester()
    app.master.title("TEH DoorBell") 
    app.master.geometry (newGeometry = "385x400+10+10")
    app.master.resizable (width="true", height="false")
    serial_init()
    write_log("TEH DoorBell started")
    app.mainloop()

if __name__ == "__main__":
  test()
